/*
   This example code is in the Public Domain (or CC0 licensed, at your option.)
   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/



/****************************************************************************
*
* This demo showcases BLE GATT client. It can scan BLE devices and connect to one device.
* Run the gatt_server demo, the client demo will automatically connect to the gatt_server demo.
* Client demo will enable gatt_server's notify after connection. The two devices will then exchange
* data.
*
****************************************************************************/

#include "main.h"
#include "helpers.h"

esp_bd_addr_t* foundedDevices;								// Содержит адреса устройств, которые уже проверенно не подходят.
static uint8_t foundedDeviceSize;
esp_bd_addr_t* myServer = NULL;								// Сервер, которые подходит.
static uint8_t probe = 0;

static const char remote_device_name[] = "Kamizelka";		// Имя сервера. Именно по этому имени клиент узнаёт сервер и устанавливает соединение чтобы поискать нужные ему сервисы.
static bool connect	= false;

static esp_ble_scan_params_t ble_scan_params = {
	.scan_type			 	= BLE_SCAN_TYPE_ACTIVE,			// Тип сканирования. Пассивный - прослушивает рекламу, активный - прослушивает и отпревляет запросы.
	.own_addr_type			= BLE_ADDR_TYPE_PUBLIC,			// Тип адресса владельца. Публичный - не может быть изменён и должен быть зарегистрирован.
	.scan_filter_policy		= BLE_SCAN_FILTER_ALLOW_ALL,	// Фильтрация.
	.scan_interval			= 0x50,							// Интервал сканирования. Time = N * 0.625 msec
	.scan_window			= 0x30,							// Окно сканирования, его длительность. Должно быть меньше или равно интервау сканирования.
	.scan_duplicate			= BLE_SCAN_DUPLICATE_DISABLE
};

/* One gatt-based profile one app_id and one gattc_if, this array will store the gattc_if returned by ESP_GATTS_REG_EVT */
static struct gattc_profile_inst gl_profile_tab[PROFILE_NUM] = {
	[PROFILE_A_APP_ID] = {
		.gattc_cb = gattc_profile_event_handler,			// Обработчик событий.
		.gattc_if = ESP_GATT_IF_NONE,	   					/* Not get the gatt_if, so initial is ESP_GATT_IF_NONE */
		.data_service_start_handle = 0,
		.data_service_end_handle = 0,
		.info_service_start_handle = 0,
		.info_service_end_handle = 0,
	},
};

static void set_bda_and_connID_profile(int prof_num, uint16_t conn_id, esp_bd_addr_t addr){
	gl_profile_tab[PROFILE_A_APP_ID].conn_id = conn_id;															// Записываем ID соединения.
	memcpy(gl_profile_tab[PROFILE_A_APP_ID].remote_bda, addr, sizeof(esp_bd_addr_t));							// Сохранение адресс подключённого устройства.
	printBoofer(GATTC_TAG, "REMOTE BDA:", gl_profile_tab[PROFILE_A_APP_ID].remote_bda, sizeof(esp_bd_addr_t));
}

static void decreaseFoundedDevices(){
	if(foundedDeviceSize < 1)
		return;
	foundedDevices = (esp_bd_addr_t*)realloc(foundedDevices, (foundedDeviceSize - 1) * sizeof(esp_bd_addr_t));	// Удаляем из списка последнее устройство.
	--foundedDeviceSize;
}

static void logDecreaseDisconnect(const char* tag, const char* str, esp_bd_addr_t bda){
	decreaseFoundedDevices();
	logDisconnect(tag, str, bda);
}

static void gattc_profile_event_handler(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param){	// Обработчик событий профиля.
	esp_ble_gattc_cb_param_t *p_data = (esp_ble_gattc_cb_param_t *)param;
	switch (event) {
	case ESP_GATTC_REG_EVT:																						// Событие после регистрации профиля.
		ESP_LOGI(GATTC_TAG, "REG_EVT");
		hCheckErr2(esp_ble_gap_set_scan_params(&ble_scan_params), "set scan params error");						// Установка параметров сканирования. Вызывает ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT.
		break;
	case ESP_GATTC_CONNECT_EVT:{																				// Событие после открытия физического соединения.
		ESP_LOGI(GATTC_TAG, "ESP_GATTC_CONNECT_EVT conn_id: %d, gattc_if: %d", p_data->connect.conn_id, gattc_if);
		set_bda_and_connID_profile(PROFILE_A_APP_ID, p_data->connect.conn_id, p_data->connect.remote_bda);		// Сохранение адресса подключённого устройства.
		probe = 0;																								// Обнуляем количество проб установки MTU.
		hCheckErr2(esp_ble_gattc_send_mtu_req (gattc_if, p_data->connect.conn_id), "config MTU error");			// Установка размера пакета MTU, определённого в esp_ble_gatt_set_local_mtu().
		// Каким - то образом наступает событие ESP_GATTC_DIS_SRVC_CMPL_EVT.
		break;	}
	case ESP_GATTC_CFG_MTU_EVT:																					// Событие успешной установки MTU.
		ESP_LOGI(GATTC_TAG, "ESP_GATTC_CFG_MTU_EVT, Status %d, MTU %d, conn_id %d", param->cfg_mtu.status, param->cfg_mtu.mtu, param->cfg_mtu.conn_id);
		if (param->cfg_mtu.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG,"config mtu failed, error status = %x. Try again.", param->cfg_mtu.status);
			if(probe < 10){
				hCheckErr2(esp_ble_gattc_send_mtu_req (gattc_if, p_data->connect.conn_id), "config MTU error");	// Установка размера пакета MTU, определённого в esp_ble_gatt_set_local_mtu().
				++probe;
			}else{
//				decreaseFoundedDevices();																		// Какие - то проблеммы с соединением, забываем про это устройство.
				logDisconnect(GATTC_TAG, "can't send MTU, remove adress, disconnected.", p_data->connect.remote_bda);
		}}
		break;
	case ESP_GATTC_OPEN_EVT:																					// Событие открытия виртуального соединения.
		if (param->open.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG, "ESP_GATTC_OPEN_EVT, open failed, status %d", p_data->open.status);
//			decreaseFoundedDevices();
			esp_ble_gap_disconnect(param->open.remote_bda);
			break;
		}else
			ESP_LOGI(GATTC_TAG, "ESP_GATTC_OPEN_EVT, open success");
		break;
	case ESP_GATTC_DIS_SRVC_CMPL_EVT:																			// Событие завершения поиска сервисов на сервере.
		ESP_LOGI(GATTC_TAG, "******** ESP_GATTC_DIS_SRVC_CMPL_EVT, discover service Start (HOW!!!!) %d", param->dis_srvc_cmpl.conn_id);
		if (param->dis_srvc_cmpl.status != ESP_GATT_OK){														// Сервер мы уже нашли во время события ESP_GAP_BLE_SCAN_RESULT_EVT и установили соединение.
			ESP_LOGE(GATTC_TAG, "discover service failed, status %d", param->dis_srvc_cmpl.status);
//			decreaseFoundedDevices();
			esp_ble_gap_disconnect(param->open.remote_bda);
			break;
		}
		ESP_LOGI(GATTC_TAG, "discover service complete conn_id %d", param->dis_srvc_cmpl.conn_id);
		esp_ble_gattc_search_service(gattc_if, param->dis_srvc_cmpl.conn_id, NULL);									// Начинаем поиск сервисов.
		break;
	case ESP_GATTC_SEARCH_RES_EVT: 																				// Событие после того, как нашли сервис на сервере.
		if (p_data->search_res.srvc_id.uuid.len == ESP_UUID_LEN_128 && compareUUID(p_data->search_res.srvc_id.uuid.uuid.uuid128, (uint8_t[]) DATA_SERVICE_UUID)) {			// Если нашли сервис данных.
			gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle = p_data->search_res.start_handle;			// Сохраняется указатель харрактеристики службы.
			gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle = p_data->search_res.end_handle;				// Сохраняется указатель харрактеристики службы.
		}else if (p_data->search_res.srvc_id.uuid.len == ESP_UUID_LEN_128 && compareUUID(p_data->search_res.srvc_id.uuid.uuid.uuid128, (uint8_t[]) INFO_SERVICE_UUID)) {	// Если нашли сервис информации.
			gl_profile_tab[PROFILE_A_APP_ID].info_service_start_handle = p_data->search_res.start_handle;			// Сохраняется указатель харрактеристики службы.
			gl_profile_tab[PROFILE_A_APP_ID].info_service_end_handle = p_data->search_res.end_handle;				// Сохраняется указатель харрактеристики службы.
		}
		break;
	case ESP_GATTC_SEARCH_CMPL_EVT:																					// Событие по завершению поиска служб.
		ESP_LOGI(GATTC_TAG, "ESP_GATTC_SEARCH_CMPL_EVT");
		if(gl_profile_tab[PROFILE_A_APP_ID].info_service_start_handle == 0 && gl_profile_tab[PROFILE_A_APP_ID].info_service_end_handle == 0){								// Если не нашли нужные нам сервисы, отключаемся.
			logDisconnect(GATTC_TAG, "Info service not found, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		if(gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle == 0 && gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle == 0){
			logDisconnect(GATTC_TAG, "Data service not found, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		uint16_t countChar = 0;
		esp_gatt_status_t status;
		status = esp_ble_gattc_get_attr_count(gattc_if, param->cfg_mtu.conn_id, 									// Получаем количество характеристик у сервиса информации.
				ESP_GATT_DB_CHARACTERISTIC, gl_profile_tab[PROFILE_A_APP_ID].info_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].info_service_end_handle, 0, &countChar);
		if(status != ESP_GATT_OK){																					// Если не получилось обнаружить харрактеристики, отключаемся с записью в логи.
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_attr_count error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		esp_gattc_char_elem_t* char_elem_result1 = (esp_gattc_char_elem_t *)malloc(sizeof(esp_gattc_char_elem_t) * countChar);		// Выделяем место для буфера значений харрактеристик с нужным UUID.
		status = esp_ble_gattc_get_char_by_uuid(gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].info_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].info_service_end_handle,
				(esp_bt_uuid_t){.len = ESP_UUID_LEN_128, .uuid.uuid128 = SERIAL_NUMBER_CHAR_UUID}, char_elem_result1, &countChar);	// Ищем харрактеристику серийного номура по UUID у этого сервиса.
		if(status != ESP_GATT_OK){
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_char_by_uuid error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		gl_profile_tab[PROFILE_A_APP_ID].serial_number_char_handle = char_elem_result1[0].char_handle;								// Сохраняем ханделы этой харрактеристики.
		status = esp_ble_gattc_get_attr_count(gattc_if, param->cfg_mtu.conn_id, ESP_GATT_DB_CHARACTERISTIC, gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle, 0, &countChar);
		if(status != ESP_GATT_OK){
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_attr_count error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		char_elem_result1 = (esp_gattc_char_elem_t *)realloc(char_elem_result1, sizeof(esp_gattc_char_elem_t) * countChar);							// Выделение места буфера для значений харрактеристик с нужным UUID.
		status = esp_ble_gattc_get_char_by_uuid(gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle,
				(esp_bt_uuid_t){.len = ESP_UUID_LEN_128, .uuid.uuid128 = LEFT_CHANNEL_CHAR_UUID}, char_elem_result1, &countChar);	// Получаем харрактеристики у сервиса.
		if(status != ESP_GATT_OK){
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_char_by_uuid error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		gl_profile_tab[PROFILE_A_APP_ID].left_channel_char_handle = char_elem_result1[0].char_handle;
		status = esp_ble_gattc_get_char_by_uuid(gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle,
				(esp_bt_uuid_t){.len = ESP_UUID_LEN_128, .uuid.uuid128 = TRANSFER_CHAR_UUID}, char_elem_result1, &countChar);
		if(status != ESP_GATT_OK){
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_char_by_uuid error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		gl_profile_tab[PROFILE_A_APP_ID].transfer_char_handle = char_elem_result1[0].char_handle;
		status = esp_ble_gattc_get_char_by_uuid(gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].data_service_start_handle, gl_profile_tab[PROFILE_A_APP_ID].data_service_end_handle,
				(esp_bt_uuid_t){.len = ESP_UUID_LEN_128, .uuid.uuid128 = DATA_CHAR_UUID}, char_elem_result1, &countChar);
		if(status != ESP_GATT_OK){
			logDisconnect(GATTC_TAG, "esp_ble_gattc_get_char_by_uuid error, disconnect", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}
		gl_profile_tab[PROFILE_A_APP_ID].data_channel_char_handle = char_elem_result1[0].char_handle;
		free(char_elem_result1);
		probe = 0;

		esp_ble_gattc_register_for_notify(gl_profile_tab[PROFILE_A_APP_ID].gattc_if,  gl_profile_tab[PROFILE_A_APP_ID].remote_bda, gl_profile_tab[PROFILE_A_APP_ID].data_channel_char_handle);

		break;
	case ESP_GATTC_REG_FOR_NOTIFY_EVT: 																			// Событие окончания регистрации на получение обновлений характеристики.

		ESP_LOGI(GATTC_TAG, "Notification registered.");

		esp_err_t ret = esp_ble_gattc_read_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].serial_number_char_handle, ESP_GATT_AUTH_REQ_NONE);
		if(ret != ESP_GATT_OK){
			ESP_LOGI(GATTC_TAG, "read char of serial number failed, status %d", ret);
//			esp_ble_gap_disconnect(gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
		}

		break;
	case ESP_GATTC_NOTIFY_EVT:																					// Событие получения уведомлений.
		// TODO обработать пришедшие данные.
		if (p_data->notify.is_notify){
			ESP_LOGI(GATTC_TAG, "ESP_GATTC_NOTIFY_EVT, receive notify value:");
		}else{
			ESP_LOGI(GATTC_TAG, "ESP_GATTC_NOTIFY_EVT, receive indicate value:");
		}
		esp_log_buffer_hex(GATTC_TAG, p_data->notify.value, p_data->notify.value_len);
		runPWM( p_data->notify.value);
		break;
	case ESP_GATTC_WRITE_DESCR_EVT:																				// Событие записи в дискриптор конфигурации клиента.
		if (p_data->write.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG, "write descr failed, error status = %x", p_data->write.status);
			break;
		}
		ESP_LOGI(GATTC_TAG, "write descr success ");
		uint8_t write_char_data[35];
		for (int i = 0; i < sizeof(write_char_data); ++i){
			write_char_data[i] = i % 256;
		}
		esp_ble_gattc_write_char( gattc_if,																		// Запись в дискриптор конфигурации клиента.
								  gl_profile_tab[PROFILE_A_APP_ID].conn_id,
								  gl_profile_tab[PROFILE_A_APP_ID].char_handle,
								  sizeof(write_char_data),
								  write_char_data,
								  ESP_GATT_WRITE_TYPE_RSP,
								  ESP_GATT_AUTH_REQ_NONE);
		break;
	case ESP_GATTC_SRVC_CHG_EVT: {																				// Событие изменения сервиса.
		esp_bd_addr_t bda;
		memcpy(bda, p_data->srvc_chg.remote_bda, sizeof(esp_bd_addr_t));
		ESP_LOGI(GATTC_TAG, "ESP_GATTC_SRVC_CHG_EVT, bd_addr:");
		esp_log_buffer_hex(GATTC_TAG, bda, sizeof(esp_bd_addr_t));
		break;
	}
	case ESP_GATTC_WRITE_CHAR_EVT:{																				// Событие завершения записи харрактеристики.
		if(p_data->write.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG, "write self address error, trying again");										// Если запись не удалась, попробуем ещё раз.
			esp_ble_gattc_write_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].left_channel_char_handle,
					ESP_BD_ADDR_LEN, esp_bt_dev_get_address(), ESP_GATT_WRITE_TYPE_NO_RSP, ESP_GATT_AUTH_REQ_NONE);
		}

		if (p_data->write.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG, "write char failed, error status = %x", p_data->write.status);
			break;
		}
		ESP_LOGI(GATTC_TAG, "write char success ");
	} break;
	case ESP_GATTC_DISCONNECT_EVT:																				// Событие закрытия физического соединения.
		connect = false;
		ESP_LOGI(GATTC_TAG, "ESP_GATTC_DISCONNECT_EVT, reason = %d", p_data->disconnect.reason);
		if(myServer != NULL){																					// Если отключились из- за проблем со связью,
			decreaseFoundedDevices();																			// Удаляем устройство из списка неподошедших.
		}
		hCheckErr2(esp_ble_gap_set_scan_params(&ble_scan_params), "set scan params filed");						// Установка параметров сканирования. Вызывает ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT.
		break;
	case ESP_GATTC_READ_CHAR_EVT:																				// Событие чтения харрактеристики.
		if(p_data->read.handle == gl_profile_tab[PROFILE_A_APP_ID].serial_number_char_handle){
			ESP_LOGI(GATTC_TAG, "Reed Serial Number");
			if(p_data->read.status != ESP_GATT_OK){
				if(probe < 10){
					ESP_LOGE(GATTC_TAG, "characteristic reading failed, error status = %x", p_data->read.status);
					esp_ble_gattc_read_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].serial_number_char_handle, ESP_GATT_AUTH_REQ_NONE);
					++probe;
				}else{
					logDisconnect(GATTC_TAG, "can't read characteristic, disconnect.", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
				}
			}
			if(p_data->read.value_len == sizeof(serialNumber) && memcmp(p_data->read.value, serialNumber, sizeof(serialNumber)) == 0 ){
				ESP_LOGE(GATTC_TAG, "Server serial number is correct");
				ESP_LOGI(GATTC_TAG, "Write my address");
				esp_ble_gattc_write_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].left_channel_char_handle,
						ESP_BD_ADDR_LEN, esp_bt_dev_get_address(), ESP_GATT_WRITE_TYPE_NO_RSP, ESP_GATT_AUTH_REQ_NONE);
			}else{
				free(myServer); // Устройство оказалось не наше, удаляем.
				myServer = NULL;
				logDisconnect(GATTC_TAG, "Server serial number is wrong, disconnect.", gl_profile_tab[PROFILE_A_APP_ID].remote_bda);
			}
		}
		if (p_data->read.status != ESP_GATT_OK){
			ESP_LOGE(GATTC_TAG, "characteristic reading failed, error status = %x", p_data->read.status);
			break;
		}
		for (int i = p_data->read.value_len - 1; i >= 0 ; --i){
			printf("%i ", p_data->read.value[i]);
		}
		printf("\n");
		break;
	default:
		printf("event uncatched: %i\n", event);
		break;
}}

static void esp_gap_cb(esp_gap_ble_cb_event_t event, esp_ble_gap_cb_param_t *param){													//Регистрация функций обработки событий gap модуля.
	uint8_t *adv_name = NULL;
	uint8_t adv_name_len = 9;
	switch (event) {
	case ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT: {																						// Событие после установки параметров сканирования.
		esp_ble_gap_start_scanning(30);																									// Начало сканирования, 30 секунд.
		break;	}
	case ESP_GAP_BLE_SCAN_START_COMPLETE_EVT:																							// Событие окончания начала сканирования.
		if (param->scan_start_cmpl.status != ESP_BT_STATUS_SUCCESS) {
			ESP_LOGE(GATTC_TAG, "scan start failed, error status = %x", param->scan_start_cmpl.status);
			break;
		}
		ESP_LOGI(GATTC_TAG, "scan start success");
		break;
	case ESP_GAP_BLE_SCAN_RESULT_EVT: {																									// Событие получения результатов сканирования.
		esp_ble_gap_cb_param_t *scan_result = (esp_ble_gap_cb_param_t *)param;
		switch (scan_result->scan_rst.search_evt) {
		case ESP_GAP_SEARCH_INQ_RES_EVT:																								// Найдено новое устройство.
			esp_log_buffer_hex(GATTC_TAG, scan_result->scan_rst.bda, 6);
			ESP_LOGI(GATTC_TAG, "searched Adv Data Len %d, Scan Response Len %d", scan_result->scan_rst.adv_data_len, scan_result->scan_rst.scan_rsp_len);
			adv_name = esp_ble_resolve_adv_data(scan_result->scan_rst.ble_adv, ESP_BLE_AD_TYPE_NAME_CMPL, &adv_name_len); 				// Получаем имя найденного устройства.
			ESP_LOGI(GATTC_TAG, "searched Device Name Len %d", adv_name_len);
			esp_log_buffer_char(GATTC_TAG, adv_name, adv_name_len);
			if(adv_name == NULL)
				break;
			if (strlen(remote_device_name) == adv_name_len && strncmp((char *)adv_name, remote_device_name, adv_name_len) == 0){		// Если имя найденного устройства совпвдвет с именем искомого сервера.
				printf("len is the same\n");
				if(myServer != NULL && !memcmp(myServer, param->scan_rst.bda, ESP_BD_ADDR_LEN)){										// Если адресс устройства совпадает с адрессом сервера, с которым уже было подключение.
					ESP_LOGI(GATTC_TAG, "Reconnect to my server");
				}else{
					for(uint8_t i = 0; i < foundedDeviceSize; ++i){																		// Проверяем, знаем ли мы уже устройство с таким же адрессом.
						if(!memcmp(foundedDevices[i], param->scan_rst.bda, ESP_BD_ADDR_LEN))											// Если знаем, то значит оно нам не подошло, то выходим.
							return;
				}}
				ESP_LOGI(GATTC_TAG, "******* searched device %s\n", remote_device_name);
				if (connect == false) {																									// Если в данный момент наше устройство не подключено ни к какому удалённому устройству.
					myServer = (esp_bd_addr_t*)realloc(myServer, sizeof(esp_bd_addr_t));												// Сохраняем аддресс устройства, как своего устройства.
					memcpy(myServer, param->scan_rst.bda, sizeof(esp_bd_addr_t));														// Если окажется, что серийные номера не навны, удалим его.
					connect = true;																										// Ставим флаг подключения.
					//				foundedDevices = (esp_bd_addr_t*)realloc(foundedDevices, (foundedDeviceSize + 1) * sizeof(esp_bd_addr_t));				// Если устройство с таким аддрессом новое.
					//				memcpy(foundedDevices[foundedDeviceSize], param->scan_rst.bda, sizeof(esp_bd_addr_t));									// Записываем аддресс.
					//				++foundedDeviceSize;
					ESP_LOGI(GATTC_TAG, "connect to the remote device.");
					esp_ble_gap_stop_scanning();																						// Останавливаем сканирование.
					// Устанавливаем соединение. Последний аргумент - установка прямого соединения или в фоновом режиме.  Наступает событие ESP_GATTC_CONNECT_EVT.
					esp_err_t err = esp_ble_gattc_open(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, scan_result->scan_rst.bda, scan_result->scan_rst.ble_addr_type, true);
					if(err){																											// Если что - то пошло не так при установке соединения, откатываем изменения.
						decreaseFoundedDevices();
						connect = false;
						err = esp_ble_gap_set_scan_params(&ble_scan_params);
						hCheckErr2(err, "set scan params filed");								// Установка параметров сканирования. Вызывает ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT.
						//esp_ble_gap_start_scanning(30);
			}}}
			break;
		case ESP_GAP_SEARCH_INQ_CMPL_EVT:																		// Событие завершения разбора результатов сканирования.
			if(connect == false)
				esp_ble_gap_start_scanning(30);
			break;
		default:
			break;
		}
		break;
	}
	case ESP_GAP_BLE_SCAN_STOP_COMPLETE_EVT:																	// Событие окончания сканирования.
		if (param->scan_stop_cmpl.status != ESP_BT_STATUS_SUCCESS){
			ESP_LOGE(GATTC_TAG, "scan stop failed, error status = %x", param->scan_stop_cmpl.status);
			break;
		}
		ESP_LOGI(GATTC_TAG, "stop scan successfully");
		break;
	case ESP_GAP_BLE_ADV_STOP_COMPLETE_EVT:
		if (param->adv_stop_cmpl.status != ESP_BT_STATUS_SUCCESS){
			ESP_LOGE(GATTC_TAG, "adv stop failed, error status = %x", param->adv_stop_cmpl.status);
			break;
		}
		ESP_LOGI(GATTC_TAG, "stop adv successfully");
		break;
	case ESP_GAP_BLE_UPDATE_CONN_PARAMS_EVT:																	// Событие завершения обновления параметров соединения.
		 ESP_LOGI(GATTC_TAG, "update connection params status = %d, min_int = %d, max_int = %d,conn_int = %d,latency = %d, timeout = %d",
				  param->update_conn_params.status, param->update_conn_params.min_int, param->update_conn_params.max_int, param->update_conn_params.conn_int,
				  param->update_conn_params.latency, param->update_conn_params.timeout);
		break;
	default:
		break;
}}

static void esp_gattc_cb(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param){	// Функция обработки событий gattc модуля.
	if (event == ESP_GATTC_REG_EVT) {																			// Событие после регистрации профиля.
		if (param->reg.status == ESP_GATT_OK) {
			gl_profile_tab[param->reg.app_id].gattc_if = gattc_if;												// Сохраняем интерфейс GATT.
		} else {
			ESP_LOGI(GATTC_TAG, "reg app failed, app_id %04x, status %d", param->reg.app_id, param->reg.status);
			return;
	}}
	do {																										// Передача события обработчику соответствующего профиля.
		for (int idx = 0; idx < PROFILE_NUM; idx++) {
			if (gattc_if == ESP_GATT_IF_NONE || gattc_if == gl_profile_tab[idx].gattc_if) {						// Если указан профиль ESP_GATT_IF_NONE, педаём событие всем профилям.
				if (gl_profile_tab[idx].gattc_cb) {
					gl_profile_tab[idx].gattc_cb(event, gattc_if, param);
	}}}} while (0);
}

void readChar (void* param){
	for( ;; ){
		esp_gatt_status_t ret = esp_ble_gattc_read_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if,
				gl_profile_tab[PROFILE_A_APP_ID].conn_id, gl_profile_tab[PROFILE_A_APP_ID].char_handle, ESP_GATT_AUTH_REQ_NONE);
		if(ret != ESP_GATT_OK)
			ESP_LOGI(GATTC_TAG, "read char failed, status %d", ret);
		else ESP_LOGI(GATTC_TAG, "read char success, status %d, char handle %i", ret, gl_profile_tab[PROFILE_A_APP_ID].char_handle);


		int r = rand() % (11);
		uint8_t d[13];
		for(int i = 0; i < 13; ++i){
			d[i] = i + r;
		}
//		esp_ble_gattc_write_char(gl_profile_tab[PROFILE_A_APP_ID].gattc_if, gl_profile_tab[PROFILE_A_APP_ID].conn_id, handleOfInterest, sizeof(d), d, ESP_GATT_WRITE_TYPE_NO_RSP, ESP_GATT_AUTH_REQ_NONE);

		vTaskDelay( 250 / portTICK_RATE_MS );
	}
	vTaskDelete( NULL );
}

void app_main(void){
	srand(42);
	pwmInitialization();
//	xTaskCreate(readChar, "TASK1", 2048, NULL, 1, NULL);
	esp_err_t ret = nvs_flash_init();											// Инициализация библиотеки энергонезависимой памяти.
	hFlashInit(ret);
	ESP_ERROR_CHECK(esp_bt_controller_mem_release(ESP_BT_MODE_CLASSIC_BT));		// Освобождение памяти и завершение работы классического режиме БТ.
	esp_bt_controller_config_t bt_cfg = BT_CONTROLLER_INIT_CONFIG_DEFAULT();	// Создание структуры БТ контроллера.
	ret = esp_bt_controller_init(&bt_cfg);										// Инициализируем и включаем БТ контроллер.
	hCheckErr(ret, "initialize controller failed", __func__);
	ret = esp_bt_controller_enable(ESP_BT_MODE_BLE);							// Устанавливаем контроллер в BLE режим.
	hCheckErr(ret, "enable controller failed", __func__);
	ret = esp_bluedroid_init();													// Инициализация Bluedroid stack (содержит API для BLE и BT).
	hCheckErr(ret, "init bluetooth failed", __func__);
	ret = esp_bluedroid_enable();												// Включение Bluedroid stack.
	hCheckErr(ret, "enable bluetooth failed", __func__);
	ret = esp_ble_gap_register_callback(esp_gap_cb);							// Регистрация функций обработки событий gap модуля.
	hCheckErr2(ret, "gap register failed");
	ret = esp_ble_gattc_register_callback(esp_gattc_cb);						// Регистрация функций обработки событий gattc модуля.
	hCheckErr2(ret, "gattc register failed");
	ret = esp_ble_gattc_app_register(PROFILE_A_APP_ID);							// Регистрация профиля.
	hCheckErr2(ret, "gattc app register failed");
	esp_err_t local_mtu_ret = esp_ble_gatt_set_local_mtu(MTU_VALUE);					// Установка размера пакета MTU (до 517).
	hCheckErr2(local_mtu_ret, "set local  MTU failed");
}
